import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

import calc.integrator.PsimuPropagationData;
import calc.integrator.firstOrderIntegrators.DopData;
import calc.main.Psimu;
import calc.main.Psimu.StorageType;
import calc.main.UserFramesConfsEnum;
import calc.output.OutputConfig;
import exception.PsimuException;
import exception.PsimuPropagationException;
import fr.cnes.genopus.nonWidget.events.CustomDateDetector;
import fr.cnes.genopus.nonWidget.exception.NonWidgetException;
import fr.cnes.genopus.nonWidget.maneuvers.CustomConstantManeuver;
import fr.cnes.genopus.nonWidget.maneuvers.CustomImpulseManeuver;
import fr.cnes.genopus.nonWidget.maneuvers.CustomManeuverSequence;
import fr.cnes.sirius.addons.patriusdataset.PatriusDataset;
import fr.cnes.sirius.patrius.assembly.Assembly;
import fr.cnes.sirius.patrius.assembly.Vehicle;
import fr.cnes.sirius.patrius.assembly.models.MassModel;
import fr.cnes.sirius.patrius.assembly.properties.MassProperty;
import fr.cnes.sirius.patrius.assembly.properties.PropulsiveProperty;
import fr.cnes.sirius.patrius.assembly.properties.TankProperty;
import fr.cnes.sirius.patrius.assembly.vehicle.VehicleSurfaceModel;
import fr.cnes.sirius.patrius.attitudes.AttitudeLaw;
import fr.cnes.sirius.patrius.attitudes.AttitudesSequence;
import fr.cnes.sirius.patrius.attitudes.LofOffset;
import fr.cnes.sirius.patrius.bodies.ExtendedOneAxisEllipsoid;
import fr.cnes.sirius.patrius.forces.ForceModel;
import fr.cnes.sirius.patrius.forces.ForceModelsData;
import fr.cnes.sirius.patrius.forces.gravity.EarthGravitationalModelFactory;
import fr.cnes.sirius.patrius.forces.gravity.EarthGravitationalModelFactory.GravityFieldNames;
import fr.cnes.sirius.patrius.frames.Frame;
import fr.cnes.sirius.patrius.frames.FramesFactory;
import fr.cnes.sirius.patrius.frames.LOFType;
import fr.cnes.sirius.patrius.math.geometry.euclidean.threed.RotationOrder;
import fr.cnes.sirius.patrius.math.geometry.euclidean.threed.Sphere;
import fr.cnes.sirius.patrius.math.geometry.euclidean.threed.Vector3D;
import fr.cnes.sirius.patrius.math.util.FastMath;
import fr.cnes.sirius.patrius.orbits.KeplerianOrbit;
import fr.cnes.sirius.patrius.orbits.PositionAngle;
import fr.cnes.sirius.patrius.propagation.MassProvider;
import fr.cnes.sirius.patrius.propagation.SpacecraftState;
import fr.cnes.sirius.patrius.propagation.events.AbstractDetector;
import fr.cnes.sirius.patrius.propagation.events.DateDetector;
import fr.cnes.sirius.patrius.propagation.events.EventDetector;
import fr.cnes.sirius.patrius.propagation.events.EventDetector.Action;
import fr.cnes.sirius.patrius.time.AbsoluteDate;
import fr.cnes.sirius.patrius.time.TimeScale;
import fr.cnes.sirius.patrius.time.TimeScalesFactory;
import fr.cnes.sirius.patrius.utils.Constants;
import fr.cnes.sirius.patrius.utils.exception.PatriusException;

/**
 * Tutorial to use PSIMU with retroplation with maneuvers and attitude sequences.
 * @author goesterjf
 *
 */
public class PsimuWithManeuversAndRetropolation {

    /**
     * @param args
     * @throws OrekitException 
     * @throws NonWidgetException 
     * @throws ParseException 
     * @throws IOException 
     * @throws PsimuException 
     * @throws PsimuPropagationException 
     */
    public static void main(String[] args) throws PatriusException, NonWidgetException, IOException, ParseException, PsimuException, PsimuPropagationException {
        
        // Patrius Dataset initialization (needed for example to get the UTC time)
        PatriusDataset.addResourcesFromPatriusDataset() ;

        // Recovery of the UTC time scale using a "factory"
        final TimeScale TUC = TimeScalesFactory.getUTC();
        
        // Reference frames
        final Frame GCRF = FramesFactory.getGCRF();
        final Frame ITRF = FramesFactory.getITRF();

        // Constants that will be used later
        final double MU   = Constants.WGS84_EARTH_MU;
        final double REQ  = Constants.WGS84_EARTH_EQUATORIAL_RADIUS;
        final double FLAT = Constants.WGS84_EARTH_FLATTENING;
        final ExtendedOneAxisEllipsoid EARTH = new ExtendedOneAxisEllipsoid(REQ, FLAT, ITRF, "EARTH");
        
        // Orbital events tuning
        final double MAXCHECK = AbstractDetector.DEFAULT_MAXCHECK;
        final double THRESHOLD = AbstractDetector.DEFAULT_THRESHOLD;
        
        // ORBIT
        
        final AbsoluteDate date = new AbsoluteDate("2010-01-01T12:00:00.000", TUC);
        
        final double sma = REQ + 250.e3;
        final double ecc = 0.;
        final double inc = FastMath.toRadians(51.6);
        final double raan = FastMath.toRadians(0.);
        final double pa  = FastMath.toRadians(0.);
        final double ano = FastMath.toRadians(0.);
        final KeplerianOrbit iniOrbit = new KeplerianOrbit(sma, ecc, inc, raan, pa, ano, PositionAngle.MEAN, GCRF, date, MU);
        
        // VEHICLE
        
        // Dry mass
        final double dryMass = 1000.;
        final MassProperty dryMassProperty = new MassProperty(dryMass);

        VehicleSurfaceModel vehicleRefSurface = new VehicleSurfaceModel(new Sphere(0.));

        // Propulsive properties
        final PropulsiveProperty engine = new PropulsiveProperty(400., 320.);
        engine.setPartName("ENGINE");
        final ArrayList<PropulsiveProperty> listOfEngines = new ArrayList<PropulsiveProperty>();
        listOfEngines.add(engine);
        final TankProperty tank = new TankProperty(200.);
        tank.setPartName("TANK");
        final ArrayList<TankProperty> listOfTanks = new ArrayList<TankProperty>();
        listOfTanks.add(tank);
        
        final Vehicle vehicle = new Vehicle(vehicleRefSurface, null, dryMassProperty, null, null, listOfEngines, listOfTanks);

        final Assembly assembly = vehicle.createAssembly(GCRF);
        final MassProvider mm = new MassModel(assembly);
        
        // FORCES
        
        final ForceModel potential = EarthGravitationalModelFactory.getBalmino(GravityFieldNames.GRGS, "grim4s4_gr", 0, 0, true);
        
        final ForceModelsData forces = new ForceModelsData(potential, null, null, null, null, null, null, null, null, null, null, null, null,  null, null);
        
        // MANEUVERS
        
        // Impulse maneuver
//        final double aol = FastMath.toRadians(180.);
//        final CustomAOLDetector eventImp = new CustomAOLDetector(aol, PositionAngle.TRUE, GCRF, 1, null, null, MAXCHECK, THRESHOLD, Action.STOP);
        final AbsoluteDate startDateImp = date.shiftedBy(-120.);
        final CustomDateDetector eventImp = new CustomDateDetector(startDateImp, MAXCHECK, THRESHOLD, Action.STOP);
        final Vector3D deltaV = new Vector3D(10., 0., 0.);
        final CustomImpulseManeuver imp = new CustomImpulseManeuver("Impulse maneuver", engine, tank, LOFType.TNW, eventImp, deltaV, mm);
        
        // Constant thrust maneuver
        final double start = -iniOrbit.getKeplerianPeriod();
        final AbsoluteDate startDate = date.shiftedBy(start);
        final CustomDateDetector startEvent = new CustomDateDetector(startDate, MAXCHECK, THRESHOLD, Action.STOP);
        final double dur = 50.;
        final AbsoluteDate endDate = startDate.shiftedBy(dur);
        final CustomDateDetector stopEvent = new CustomDateDetector(endDate, MAXCHECK, THRESHOLD, Action.STOP);
        final Vector3D direction = new Vector3D(-1., 0., 0.);
        CustomConstantManeuver cont = new CustomConstantManeuver("Continuous maneuver", engine, tank, LOFType.TNW, startEvent, stopEvent, direction, mm);
        
        final CustomManeuverSequence manSeq = new CustomManeuverSequence(0., 0.);
        manSeq.addManeuver(imp);
        manSeq.addManeuver(cont);        
        
        // ATTITUDE (with a switch but not at the same date as initial conditions !)
        
        final AttitudeLaw law1 = new LofOffset(LOFType.TNW, RotationOrder.ZYX, 0., 0., 0.);
        // Switch time must not be exactly at the same date as inital conditions (Patrius bug).
        final EventDetector event = new DateDetector(date.shiftedBy(1.0), MAXCHECK, THRESHOLD, Action.RESET_STATE);
        final AttitudesSequence seqAtt = new AttitudesSequence();
        seqAtt.addSwitchingCondition(law1, event, true, false, law1);
                
        // INTEGRATION DATA

        final double duration = -86400.;
        final double hStop = 120.e+3;
        final PsimuPropagationData propagationData = new PsimuPropagationData(duration, GCRF, hStop);
        
        final double minStep = 0.1;
        final double maxStep = 30.;
        final double[] vecAbsoluteTolerance = { 1.0E-5, 1.0E-5, 1.0E-5, 1.0E-8, 1.0E-8, 1.0E-8 };
        final double[] vecRelativeTolerance = { 1.0E-8, 1.0E-8, 1.0E-8, 1.0E-8, 1.0E-8, 1.0E-8 };
        final double absMassTolerance = 1.0E-3;
        final double relMassTolerance = 1.0E-2;
        final DopData integrationData = new DopData(propagationData, minStep, maxStep, vecAbsoluteTolerance, vecRelativeTolerance, absMassTolerance, relMassTolerance);
        
        final Psimu test = new Psimu(iniOrbit, EARTH, UserFramesConfsEnum.IGNORE, integrationData, vehicle, mm, forces, manSeq, seqAtt);
        
        System.out.println("Initial orbit:");
        System.out.println("Date:"+iniOrbit.getDate());
        System.out.println("SMA:"+iniOrbit.getA());
        System.out.println("MASS:"+mm.getTotalMass()+" kg");

        // PROPAGATION
        
        final OutputConfig output = new OutputConfig(FramesFactory.getEME2000(), LOFType.LVLH, 3600.);        
        test.propagateInMasterMode(output, null, StorageType.MEMORY);
      
        List<SpacecraftState> listOfSc = test.getSpacecraftStateList();
      
        final int nb = listOfSc.size();
        
        System.out.println();
        System.out.println("Final orbit:");
        System.out.println("Date:"+listOfSc.get(0).getDate());
        System.out.println("SMA:"+listOfSc.get(0).getA()/1000.+" km");
        System.out.println("MASS:"+mm.getTotalMass()+" kg");
        
        System.out.println("\nAmount of S/C states = "+nb);
        for (SpacecraftState spacecraftState : listOfSc) {
            System.out.println(spacecraftState.getDate());
        }
        
        test.saveVTSFiles("OUTPUT_JPSIMU", "testManRetro");

    }
    
}
